	list p=16f628a
	
; PILbox1.asm
; PIL-Box version 1 - HPIL/Serial Translator based on PIC uP
; History:
; 28/10/07 prototype code for feasability
; 13/01/08 test with  PIC internal comparators
; 08/03/08 translator implementation
; 25/10/08 some cleanup and debugging
; 28/11/08 version 1.0: more cleanup, documentation
; 08/12/08 version 1.1: fix problem with HP-71B RESTORE IO sequence


	#include <P16F628A.INC>
;
; config bits: HS osc, watchdog off, powerup timer on, code protect off,
; brownout enabled, master clear disabled, low voltage prog disabled, data ee protect disabled.

        __config 0x2142   ; 16F628A


;
; 20-6f: bank of 80 bytes
BUF     equ     0x20            ; buffer for IL frame samples (20-5f)
DECFLG  equ     0x60		; decoding flags
CURIN   equ     0x61		; current input sample
CPTBIT  equ     0x62		; bit counter
CPTSAM  equ     0x63		; sample counter
FRREGH  equ     0x64		; input/output frame shift registers
FRREGL  equ     0x65
; 70-7f: commun area 
ILFLG   equ     0x70            ; IL flags 
FRREC   equ     0      ; frame received
FRAV    equ     1      ; frame available  
FCMD    equ     2      ; command frame received
CPTB    equ     0x71            ; general purpose counter
FRAMEH  equ     0x72            ; IL 11-bit frame in 72-73
FRAMEL  equ     0x73
SBYTE   equ     0x74           ; byte from/to serial
HBYT    equ     0x75           ; high/low frame bytes from/to serial
LBYT    equ     0x76
LASTH   equ     0x77           ; last serial bytes
LASTL   equ     0x78
LASTC   equ     0x79           ; last command frame


LEDOFF  equ     4      ; LED command - RA4

;
	org     0
	goto    Start

        org     4
Intr
; interrupt routine to sample the HPIL frame
     variable i
i=0
     while i<D'64'   ; 128 samples stored in 64 registers
        movf    CMCON,w     ; ABxx11xx
        movwf   BUF+i
        swapf   CMCON,w     ; 11xxCDxx
        andwf   BUF+i,f     ; ABxxCDxx
i+=1
     endw
        bcf     PIR1, CMIF   ; reset IT cause
        bsf     ILFLG,FRREC
        retfie

; -----------------------------------------------
        dt "J-F GARNIER, 2008"
; -----------------------------------------------

Start
	movlw   B'00000000'     ; port outputs
	movwf   PORTA
	movwf   PORTB
      bsf     STATUS, RP0     ; *****************
	movlw   B'11101111'     ;make port a4 output
	movwf   TRISA           ;       /
	movlw   B'11010010'     ;make port b0,2,3,5 outputs
	movwf   TRISB           ;       /
      bcf     OPTION_REG,7    ; weak pull-up on port B
      call    Dly100us	      ; delay 100us

; -- setup of the UART
        movlw   D'129'       ; 9600 bauds
        movwf   SPBRG
; setup Rx and Tx
        movlw   0x24         ; Tx enable, async high speed
        movwf   TXSTA        
        bcf     STATUS, RP0  ; *************************
        movlw   0x90         ; Serial enable, continous receive
        movwf   RCSTA        

; -- setup of the comparators
        movlw   0x1c         ;  mode 100, C1INV
        movwf   CMCON
        nop
        movf    CMCON,w       ; read

; enable INT on comparator change
        bsf     STATUS, RP0  ; *************************
        movlw   0x40            ; CMIE
        movwf   PIE1
        bcf     STATUS, RP0  ; *************************
        movlw   0x40            ; PEIE
        movwf   INTCON

; init variables        
        clrf    ILFLG
        clrf    LASTH
        clrf    LASTL
        
        
; ======== main loop ===========
Main
        bsf     PORTA,LEDOFF   ; LED off
        bsf     INTCON, GIE ; enable intr
; internal loop, waiting for IL frame, or serial byte
Loop    btfsc   ILFLG,FRREC
        goto    insamp
        btfsc   ILFLG,FRAV
        goto    inIL
        btfsc   PIR1,RCIF
        goto    inserial
        goto    Loop
; ================================



; -------------------------------        
; IL frame samples received.
; -------------------------------
insamp
        bcf     INTCON, GIE   ; disable intr
        bcf     PORTA,LEDOFF       ; LED on
        bcf     ILFLG,FRREC
        call    decframe      ; decode frame into FRAMEH/L,  FRAV is set if ok
	goto    Main
        
; -------------------------------        
; IL frame available.
; -------------------------------
inIL    bcf     INTCON, GIE   ; disable intr
        bcf     PORTA,LEDOFF       ; LED on
        bcf     ILFLG,FRAV
        btfss   FRAMEH,2   ; control frame?
        goto    TR12       ;   no (DOE), translates it
        btfsc   FRAMEH,1   ; CMD or RDY?
        goto    TR11       ;   no (IDY), retransmit
        btfss   FRAMEH,0   ; RDY?
        goto    TR10       ;   no (CMD)
        movf    FRAMEL,w   ; 
        btfss   STATUS,Z   ; RFC?
        goto    TR12       ;   no, translates it
        ; RFC frames
        btfss   ILFLG,FCMD
        goto    Main
        bcf     ILFLG,FCMD
        ; translate previously received command
        movf    LASTC,w
        movwf   FRAMEL
        movlw   0x04
        movwf   FRAMEH
        goto    TR12

TR10    ; command: store and retransmit
	  movf    FRAMEL,w
        movwf   LASTC
        bsf     ILFLG,FCMD
TR11    call    sndframe
        goto    Main


        
TR12                
        ; build low and high parts
        movf    FRAMEL,w
        movwf   LBYT
        movf    FRAMEH,w
        movwf   HBYT
        ; test for last format used for low byte:
	btfss   LASTL,7
        goto    TR13
        ; 1ddd dddd          
        rlf     LBYT,w
        rlf     HBYT    ; 4 higher bits
        rlf     HBYT,w
        andlw   0x1e
        iorlw   0x20    ; format = 001c ccd0
        movwf   HBYT
        bsf     LBYT,7  ; format = 1ddd dddd
        goto    TR14
        
TR13    ; 01dd dddd
        rlf     LBYT
        rlf     HBYT
        rlf     LBYT,w
        rlf     HBYT,w    ; 5 higher bits
        andlw   0x1f
        iorlw   0x20    ; format = 001c ccdd
        movwf   HBYT
        rrf     LBYT
        bcf     LBYT,7
        bsf     LBYT,6  ; format = 01dd dddd
        
TR14    ; if high part different than last one, send it
        movf    LASTH,w
        subwf   HBYT,w
        btfsc   STATUS,Z
        goto    TR15
        movf    HBYT,w
        movwf   LASTH
        call    sndserial
        goto    TR16
        
TR15    ; else send low part
        movf    LBYT,w
        movwf   LASTL
        call    sndserial
        
TR16    goto Main    
        
        
       

; -------------------------------        
; serial byte received.
; -------------------------------        
inserial
        bcf     INTCON, GIE    ; desable intr
        bcf     PORTA,LEDOFF      ; LED on
        movf    RCREG,W
        movwf   SBYTE
        btfss   RCSTA,OERR
        goto    inSER1
; in case of USART error, reset it
        bcf     RCSTA,CREN
        nop
        bsf     RCSTA,CREN        
inSER1         
        ; check for high byte ack
        movf    SBYTE,w
        sublw   0x0D  ; CR?
        btfsc   STATUS,Z
        goto    TR20
        movf    SBYTE,w
        sublw   0x0A  ; LF?
        btfss   STATUS,Z
        goto    TR22
TR20    ; send low byte
	movf    LBYT,w
	btfsc   STATUS,Z   ; already sent?
	goto    TR21       ; yes
	movwf   LASTL
        call    sndserial
        clrf    LBYT       ; sent!
TR21    goto    Main

TR22  ; test for low/high byte
	btfsc   SBYTE,7
	goto    TR23   ; low byte
	btfsc   SBYTE,6
	goto    TR23   ; low byte
      btfss   SBYTE,5
      goto    Main   ; invalid
	; high byte, store it
	movf    SBYTE,w
	movwf   HBYT
	goto    Main

TR23    ; else low byte
	movf    SBYTE,w
	movwf   LASTL
        ; was a high byte received?
        movf    HBYT,w
        btfsc   STATUS,Z
        goto    TR28  ; no
        ; yes, use it
        movwf   LASTH
        clrf    HBYT
TR28    ; assemble frame
        movf    LASTL,w
        movwf   FRAMEL
        movf    LASTH,w
        movwf   FRAMEH
	; test format
        btfss   LASTL,7
	goto    TR24
	; low = 1bbb bbbb
	rlf     FRAMEL
	rrf     FRAMEH
	rrf     FRAMEH
	rrf     FRAMEL
	goto    TR25
TR24    ; low = 01bb bbbb
	rlf     FRAMEL
	rlf     FRAMEL
	rrf     FRAMEH
	rrf     FRAMEL
	rrf     FRAMEH
	rrf     FRAMEL
TR25    movf    FRAMEH,w
	andlw   7
	movwf   FRAMEH

        ; test for special frames
	movf	FRAMEH,w
	sublw	4
	btfss	STATUS,Z
	goto	TR30
      ; command frames
	movf	FRAMEL,w
      andlw 0xf4
      sublw 0x94
	btfss	STATUS,Z
	goto	TR26
	; send back ack to PC
	movf	LASTL,w
	call	sndserial
	goto	Main

TR26  ; send RFC
      movlw  0x05
      movwf  FRAMEH
      clrf   FRAMEL
        
TR30     ; transmit it
        movf    CMCON,w  
        bcf     PIR1, CMIF   ; clear comparator IT
	call    sndframe        
        btfss   PIR1, CMIF   ; check comparator IT
        goto    Main         ; ok, return

        movf    CMCON,w  
        bcf     PIR1, CMIF   ; clear comparator IT
	bsf     ILFLG,FRAV   ; frame available
	goto    Main 
	
                

; ---------------------------------	
        
        
        
; send a byte to serial
sndserial
        btfss   PIR1,TXIF
        goto    sndserial
        movwf   TXREG
        return
        
		
; ---------------------------------	

decframe
; decode frame from samples in BUF (20-5f). Result in FRAME, set ILFLG,FRAV if frame found
        movlw   BUF
        movwf   FSR
        movlw   40     ; 64 registers (128 samples)
        movwf   CPTSAM
        clrf    CPTBIT
        clrf    DECFLG ; flags for decoding
; DECFLG flags: 
fBIT    equ     0    ; fBIT must be bit 0
fSTART  equ     1
fGETB   equ     2
fIDLE   equ     3
; CURIN bit positions:
fINp    equ     6
fINm    equ     7

decfra1
; get input samples in CURIN
        movf   INDF,W
        movwf  CURIN
        incf   FSR
        bcf    DECFLG,fGETB
        goto   getsam3

getsam2 
        swapf  CURIN
        bsf    DECFLG,fGETB
getsam3
; find bits
; positive pulse? 
        btfss   CURIN,fINp
        goto    fndtr1
        bsf     DECFLG,fSTART
        bcf     DECFLG,fBIT    ; last positive pulse will mean '0' bit
        bcf     DECFLG,fIDLE
        goto    fndtr4
fndtr1
; negative pulse?
        btfsc   CURIN,fINm
        goto    fndtr2
        bsf     DECFLG,fSTART
        bsf     DECFLG,fBIT    ; last negative pulse will mean '1' bit
        bcf     DECFLG,fIDLE
        goto    fndtr4
fndtr2
; check IDLE state after at least one pulse is found
        btfss   DECFLG,fSTART
        goto    fndtr4
; idle state found
        btfsc   DECFLG,fIDLE
        goto    fndtr5
        bsf     DECFLG,fIDLE
        goto    fndtr4
fndtr5
; idle found 2 times: one bit found, store it
        rrf     DECFLG,W
        rlf     FRREGL
        rlf     FRREGH
        incf    CPTBIT
        bcf     DECFLG,fSTART        
        bcf     DECFLG,fIDLE

fndtr4
        btfss  DECFLG,fGETB
        goto   getsam2

        decfsz  CPTSAM
        goto    decfra1

; end of samples. did we get exactly 11 bits?
        movf    CPTBIT,W
        sublw   D'11'
        btfss   STATUS,Z
        goto    fndtr9              ; no
        bsf     ILFLG,FRAV          ; yes
        movf    FRREGH,w
        andlw   7
        movwf   FRAMEH
        movf    FRREGL,w
        movwf   FRAMEL
fndtr9  return
        


; ------------------------------------------------------------------------        
        
sndframe
; send a frame from FRAME
MASKM   equ  1
MASKP   equ  8
        movf  FRAMEH,w
	  movwf FRREGH
        movf  FRAMEL,w
	  movwf FRREGL
        movlw D'11'   ; 11 bits
        movwf CPTBIT

        btfsc FRREGH,2
        goto  sendf1

; send '0' sync bit
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKP  ; OUT+
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKP  ; OUT+
        movwf  PORTB
        goto   sendf2

sendf1
; send '1' sync bit
        movlw  MASKP  ; OUT+
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKP  ; OUT+
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
sendf2  nop
        movlw  0  ; 
        movwf  PORTB
        nop
        nop
       
        decf   CPTBIT   ; is now 10
              
sendn       
        rlf   FRREGL
        rlf   FRREGH
        btfsc FRREGH,2
        goto  send1
        nop
; send '0' bit
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKP  ; OUT+
        movwf  PORTB
        goto send2

send1
; send '1' bit
        movlw  MASKP  ; OUT+
        movwf  PORTB
        nop
        nop
        nop
        movlw  MASKM  ; OUT-
        movwf  PORTB
        nop
        nop
send2   nop
        movlw  0  ; 
        movwf  PORTB

        decfsz CPTBIT
        goto sendn
        return



; delay of W * 100 us
Delay	
	movwf	CPTB
d100ms call	Dly100us
	decf	CPTB,F
	btfss	STATUS,Z
	goto	d100ms
	return


Dly100us			; delay 100us = 125 loops of 0.8us
	movlw	D'131'		; (from 131 to 256)
d1ms	addlw	1		; 1 cycle (0.2 us at 20MHz)
	btfss	STATUS,Z	; 1
	goto	d1ms		; 2
	return


	end

	
